import logging

import click
import click_log

from ib_insync import *
from .ibkrbox import *

logger = logging.getLogger(__name__)
click_log.basic_config(logger)

CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"], auto_envvar_prefix="IBKRBOX"
)


@click.command(context_settings=CONTEXT_SETTINGS)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-i",
    "--ip",
    default="localhost",
    help="ip address or hostname of IBKR tws or gateway",
    type=str,
)
@click.option(
    "-p", "--port", default=7496, help="port of IBKR tws or gateway", type=int
)
@click.option("--s1", help="first strike", type=float)
@click.option("--s2", help="second strike", type=float)
@click.option("-r", "--rate", help="rate to use", type=float)
@click.option("-a", "--amount", help="amount per contract", type=float)
@click.option("-m", "--months", help="duration in months", type=int)
@click.option("-e", "--expiry", help="contract expiry to use", type=str)
@click.option("--acc", default="", help="account id to use", type=str)
@click.option("-l", "--limit", help="limit price", type=float)
@click.option("-t", "--timeout", default=20, help="seconds to wait for order fill", type=int)
@click.option("-l", "--limit", help="limit price", type=float)
@click.option("--offset", help="will try until price limit+-offset", default=None, type=float)
@click.option("-q", "--quantity", default=1, help="num contracts", type=int)
@click.option("--short", help="sell the box to loan", is_flag=True)
@click.option("--es", help="use ES options", is_flag=True)
@click.option("--execute", help="send order for execution", is_flag=True)
def cli(
    ip, port, s1, s2, rate, amount, months, expiry, limit, quantity, short, es, execute, acc,timeout,offset
):
    """Construct a Box Spread using SPX or ES options.
    Use "execute" option to send order for execution.
    """
    assert months != None or limit != None, "pass limit price, or months to calc limit"
    assert months != None or expiry != None, "pass expiry, or months to calc expiry"
    assert amount != None or (
        s1 != None and s2 != None
    ), "pass strikes or amount per contract"
    ib = get_ib(ip, port, acc)
    assert ib.isConnected(), "IB connection error, please retry"
    if not expiry:
        expiry = get_expiry_es(ib, months) if es else get_expiry(ib, months)
    if s1 and not s2:
        s2 = s1 + int(amount / (50.0 if es else 100.0))
    if s2 and not s1:
        s1 = s2 - int(amount / (50.0 if es else 100.0))
    if not s1:
        s1, s2 = get_strikes(ib, amount, es)
    if not limit:
        if not rate:
            rate = get_rate(expiry) + 0.30
        limit = get_limit(expiry, rate, s1, s2, es)
    limit = -abs(limit) if short else abs(limit)
    if not offset:
        max = limit
    else:
        assert offset/limit < .05, "offset should be with in spread, or less than 5%"
        max = limit + offset
    mintick = 0.05
    limit, max = mintick * round(limit / mintick), mintick * round(max / mintick)
    print(f"\nUsing expiry, s1, s2, price limit/max, quantity, short, es: {expiry}, {s1}, {s2}, {limit:.2f}/{max}, {quantity}, {short}, {es}"
        )
    if rate: print(f'using interest rate: {rate:.2f}')
    trade = box_trade(ib, expiry, s1, s2, limit, quantity, short, es, not execute, acc, timeout, max)
    if trade != None:
        ib.sleep(5)



from datetime import datetime
import sys
import argparse
import pickle
import pandas as pd
from ib_insync import *
import platform
import numpy as np
import math
import tempfile
from pathlib import Path

def get_ib(ip, port, acc=""):
    ib = IB()
    ib.connect(ip, port, clientId=19, timeout=20, account=acc)
    return ib

def get_last(ib, con, dur="3600 S", bar="1 min", rth = False):
    ib.reqMarketDataType(3)
    try:
        bars = ib.reqHistoricalData(
            con,
            endDateTime="",
            durationStr=dur,
            barSizeSetting=bar,
            whatToShow="TRADES",
            useRTH=rth,
        )
    except:
        bars = []
    if len(bars) == 0:
        last = math.nan
        if "D" not in dur:
            last = get_last(ib, con, dur="4 D", bar="1 day")
        return last
    return bars[-1].close

def box_trade(
    ib,
    expiry,
    strike1,
    strike2,
    limit,
    quantity=1,
    short=False,
    is_future=False,
    show=True,
    acc = "",
    timeout = 10,
    max=None
):
    assert strike1 < strike2, "incorrect strikes"
    sym = "ES" if is_future else "SPX"
    exchange = "CME" if is_future else "SMART"
    mul = 50 if is_future else 100

    if not short:
        boxorder = ["BUY", "SELL", "SELL", "BUY"]
        assert limit > 0.0
        assert max >= limit
    else:
        boxorder = ["SELL", "BUY", "BUY", "SELL"]
        assert limit < 0.0
        assert max >= limit

    boxspread = [
        FuturesOption(sym, expiry, S, T, exchange=exchange, tradingClass="EW")
        if is_future
        else Option(sym, expiry, S, T, exchange=exchange, tradingClass="SPX")
        for S in (strike1, strike2)
        for T in ("C", "P")
    ]
    boxspread = ib.qualifyContracts(*boxspread)

    legs = [
        ComboLeg(conId=x[0].conId, ratio=1, action=x[1], exchange=exchange)
        for x in zip(boxspread, boxorder)
    ]
    bag = Contract(
        symbol=sym, secType="BAG", exchange=exchange, currency="USD", comboLegs=legs
    )
    _df = util.df(boxspread)[
        ["symbol", "strike", "right", "lastTradeDateOrContractMonth"]
    ]
    _df.loc[:, "action"] = util.df(bag.comboLegs)["action"]
    print(f"\nBox Legs:\n{_df}")
    if not short:
        print(
            f"Lend {quantity*int(limit*mul)} today, receive {quantity*(strike2-strike1)*mul} on {expiry}"
        )
    else:
        print(
            f"Borrow {abs(quantity*int(limit*mul))} today, repay {quantity*(strike2-strike1)*mul} on {expiry}"
        )
    if show:
        return None

    trade = None
    mintick = .05
    for limit in np.arange(limit, max+mintick, mintick):
        if quantity == 0: break
        print(f'placing limit order for {quantity} at {limit:.2f}. Waiting for {timeout}s...')
        trade = ib.placeOrder(bag, LimitOrder("BUY", quantity, limit, account=acc))
        try:
            ib.sleep(timeout)
        except:
            ib.cancelOrder(trade.order)
            return None
        if trade.filled() != 0:
            print(f'order filled for {trade.filled()} at {trade.orderStatus.avgFillPrice:.2f}')
        if trade.filled() != quantity:
            print(f'cancelling unfilled order, remaining quantity {quantity}')
            ib.cancelOrder(trade.order)
            ib.sleep(3)
        else:
            break
        quantity -= trade.filled()
    return trade


def get_rate(expiry, show=True):
    days = (datetime.strptime(expiry, "%Y%m%d") - datetime.now()).days
    # pull rates from treasury
    assert days < 7 * 365
    dt = datetime.now().strftime("%Y%m%d")
    tmpf = (
        Path("/tmp" if platform.system() == "Darwin" else tempfile.gettempdir())
        / f"{dt}.rates"
    )
    try:
        rates_ = pd.read_pickle(tmpf)
        rates = list(rates_.iloc[0])[1:11]
    except:
        rates = None
    if not rates:
        url_ = f'https://home.treasury.gov/resource-center/data-chart-center/interest-rates/daily-treasury-rates.csv/all/{datetime.now().strftime("%Y%m")}?type=daily_treasury_yield_curve&field_tdr_date_value_month={datetime.now().strftime("%Y%m")}&page&_format=csv'
        # print(url_)
        rates_ = pd.read_csv(url_)
        assert list(rates_.columns)[1:11] == [
            "1 Mo",
            "2 Mo",
            "3 Mo",
            "4 Mo",
            "6 Mo",
            "1 Yr",
            "2 Yr",
            "3 Yr",
            "5 Yr",
            "7 Yr",
        ], "problem with treasury rates site, pass rate or limit price manually"
        rates = list(rates_.iloc[0])[1:11]
        rates_.to_pickle(tmpf)
    assert rates != None, "problem with rate calculation, pass rate or limit manually"
    rate = rates[
        [
            n
            for n, i in enumerate(
                [30, 60, 90, 120, 180, 365 * 1, 365 * 2, 365 * 3, 365 * 5, 365 * 7]
            )
            if i > days
        ][0]
    ]
    if show:
        print(f"selected treasury rate: {rate}\n{rates_}")
    return rate


def get_expiry_es(ib, months):
    exp = [
        x.realExpirationDate
        for x in ib.reqContractDetails(Future("ES", exchange="CME"))
    ]
    exp.sort()
    exp = exp[: int(months / 3) + 2]
    for i in exp:
        spx = Future("ES", i, "CME")
        spx = ib.qualifyContracts(spx)[0]
        chains = ib.reqSecDefOptParams(spx.symbol, "CME", spx.secType, spx.conId)
        chain = next(c for c in chains if c.exchange == "CME")
        exps = [x.expirations for x in chains if x.tradingClass == "EW"][0]
        if months < 1:
            return exps[0]
        for j in exps:
            if (datetime.strptime(j, "%Y%m%d") - datetime.now()).days >= months * 30:
                return j
    assert False, "expiry not found"


def get_expiry(ib, months):
    spx = ib.qualifyContracts(Index("SPX", "CBOE"))[0]
    chain = next(
        c
        for c in ib.reqSecDefOptParams(spx.symbol, "", spx.secType, spx.conId)
        if c.tradingClass == "SPX" and c.exchange == "SMART"
    )
    expdays = [
        abs(months * 30 - (datetime.strptime(exp, "%Y%m%d") - datetime.now()).days)
        for exp in chain.expirations
    ]
    _, idx = min((v, i) for i, v in enumerate(expdays))
    return chain.expirations[idx]


def get_strikes(ib, amount, is_futures):
    spx = ib.qualifyContracts(Index("SPX", "CBOE"))[0]
    price = get_last(ib, spx)
    print(f"SPX last close: {price}")
    assert not util.isNan(price) and price > 100, "invalid SPX market price"
    spread = int(amount / (50.0 if is_futures else 100.0))
    assert spread % 10 == 0, "use amount in 1000s"
    strike = int(price / 100) * 100
    if spread <= 200:
        return strike, strike + spread
    else:
        rem = spread % 200
        remhalf = (spread - rem) / 2
        return strike - remhalf, strike + remhalf + rem


def get_limit(expiry, rate, s1, s2, is_future=False):
    days = (datetime.strptime(expiry, "%Y%m%d") - datetime.now()).days
    if not rate:
        rate = get_rate(expiry) + 0.30
    limit = (s2 - s1) / (1 + rate / 100.0) ** (days / 365)
    mintick = 0.05
    limit = mintick * round(limit / mintick)
    return limit



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ibkrbox']

package_data = \
{'': ['*']}

install_requires = \
['click-log>=0.4.0,<0.5.0',
 'click>=8.1.3,<9.0.0',
 'ib-insync>=0.9.71,<0.10.0',
 'pandas>=1.5.1,<2.0.0']

entry_points = \
{'console_scripts': ['ibkrbox = ibkrbox.cli:cli']}

setup_kwargs = {
    'name': 'ibkrbox',
    'version': '0.6.0',
    'description': 'box spread utility for interactive brokers',
    'long_description': '# ibkrbox\nConstructs a Box Spread combo order for SPX or ES futures option, only required arguments are amount you want to lend or borrow , and for how many months.\n\nThis utility will automatically look up current treasury rates, and add .30 to get the yield rate. This will be used to calculate limit price. rate or limit price can be overridden as needed.\n\nIt can also automatically calculate the right strikes and spread, with approximate expiry for given duration. All of these can be overridden as needed.\n\nThis utility is easy to install and use with existing IBKR TWS or gateway session. Just make sure to enable API access in the GUI of IBKR TWS or gateway.\n\nPlease file a issue if you notice any problem(s).\n\n## Installation\n```code\npip install ibkrbox\n```\n\n## Usage\n\n```code\nibkrbox -h\n```\n<img width="630" alt="image" src="https://user-images.githubusercontent.com/998264/215016906-f72926c9-bada-4430-a2bb-c07db5ea69c6.png">\n\n\n### 1. construct a combo SPX Box Spread lending for 50K, duration 4 months (use "--execute" option to send the order to IBKR)\nThis will not execute the order, so you can safely run this.\n```code\nibkrbox -a 50000 -m 4\n```\n<img width="474" alt="image" src="https://user-images.githubusercontent.com/998264/215017182-3577e49f-7787-41c9-b500-303b6afded19.png">\n\n\n### 2. same as above but using Options on ES Futures (use "--execute" option to send the order to IBKR)\nThis will not execute the order, so you can safely run this.\n```code\nibkrbox -a 50000 -m 4 --es\n```\n<img width="469" alt="image" src="https://user-images.githubusercontent.com/998264/215017485-1fb9cd8c-bf0c-44e8-8775-7844831a8f85.png">\n\n\n\n### 3. construct a combo SPX Box Spread borrowing for 50K, duration 4 months (use "--execute" option to send the order to IBKR) \nThis will not execute the order, so you can safely run this.\n```code\nibkrbox -a 50000 -m 4 --short\n```\n',
    'author': 'asemx',
    'author_email': '998264+asemx@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/asemx/ibkrbox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
